/* Example DevMUD module to exercise interface standard and socket module */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include "../include/devmud.h"

/* defines */

#define SNEAKY_FLAG 1
#define INVIS_FLAG 2

#define O_INVALID	(-1)
#define O_NONE		0
#define O_ARG		1
#define O_HIDDEN	2
#define O_PRIV		4
#define O_NOTPRIV	0

#define READ_BUFFER_SIZE 16384
#define PASTE_BUFFER_SIZE READ_BUFFER_SIZE
#define OUT_BUFFER_SIZE (READ_BUFFER_SIZE+256)

#define CLASS_SYSTEM 1
#define CLASS_OTHER 2
#define CLASS_MOB 4
#define CLASS_SELF 8

/* some types */
enum types { NOTYPE, STRING, NUMBER };

struct node {
	/* hash portion */
	int num_buckets, count, collisions;
	struct node **all_buckets;
	/* bucket portion, key */
	unsigned int hashvalue;	/* for approx ID check and resize */
	enum types keytype;
	union {
		char *text;	/* malloc'd for exact ID check */
		int val;
	} key;
	/* bucket portion, data */
	enum types type;
	union {
		char *text;	/* malloc'd, of course */
		int num;
	} u;
	/* bucket next, up */
	struct node *next, *parent;
};

struct playerinfo {
	/* function to send messages */
	void (*message)(int id,char *message,int len,int oob);
	int id;	/* network ID */
	char command_buffer[READ_BUFFER_SIZE];
	char paste_buffer[READ_BUFFER_SIZE];
	int command_tail;
	int state;
	int wrap;
	struct node *object;
	int timezone;
	int timestamp;
};

struct mobinfo {
	struct node *object;
	int delay;
};

/* function pointers to things this module uses */
static void (*queue_event)(struct timeval when, void (*func)(struct timeval when, int data), int data) = NULL;
static void (*close_connection)(int which) = NULL;
static int (*real_connection)(int which) = NULL;
static char *(*connection_name)(int which) = NULL;
static struct timeval (*connection_time)(int which) = NULL;
static int (*open_port)
		(int port,
		 unsigned int addr,
		 int (*new_connection)
		       (int id,
			void (*message)(int id,char *message,int len,int oob)),
		 void (*end_connection)(int id),
		 void (*message)(int id,char *message,int len,int oob))
	= NULL;
static void (*close_port)(int id) = NULL;

/* forward declaractions */
static void cleanup_data(struct node *it);
static void print_data(FILE *fp, struct node *it);
static void read_data(FILE *fp, struct node *it);

static void init_assoc(struct node *h, int buckets);
static void uninit_assoc(struct node *h);
static void cleanup_assoc(struct node *h);
static struct node *new_string_node(struct node *h, char *key);
static struct node *new_number_node(struct node *h, int key);
static struct node *find_string_node(struct node *h, char *key);
static struct node *find_number_node(struct node *h, int key);
static struct node *find_keyed_node(struct node *h, struct node *key);
static int resize_assoc(struct node *h, int buckets);
static void print_hash(struct node *h, FILE *fp, int level);
static void read_hash(struct node *h, FILE *fp);

int bigworld_stop(int severity);
static void send_string(int who, char *str, int start, int class);

/* module-global variables */

static struct node db, *objects, *auth, *obcount, *root_object, *start_room;
static struct playerinfo *p;
static struct mobinfo *m;
static int port_number = 2121, max_players = 200, max_mobs = 8;
/* max_players and max_mobs are current array sizes */

/* functions */

/* out of memory or some such */
static void panic(void)
{
	fprintf(stderr,"Panic quit\n");
	print_hash(&db,stderr,0);
	bigworld_stop(2);
}

static void *s_malloc(size_t size)
{
	void *mem;

	mem = malloc(size);
	if (mem == NULL)
	{
		fprintf(stderr,"Fatal: Out of memory.\n");
		panic();
		exit(-1);
	} else {
		return(mem);
	}
}

static char *s_strdup(char *s)
{
	void *str;

	str = strdup(s);
	if (str == NULL)
	{
		fprintf(stderr,"Fatal: Out of memory.\n");
		panic();
		exit(-1);
	} else {
		return(str);
	}
}

/* functions for referring to structures by strings (associative array) */

/* take a string and return the hash */
/* does not depend on which hash table is involved */
static unsigned int hash_key(char *key)
{
	unsigned int temp;

	temp = 0;
	while(*key)
	{
		temp = temp*43 + *key++;
	}
	return(temp);
}

struct traverse {
	struct node *h;		/* what I'm traversing */
	int index;		/* index into all_buckets */
	struct node *b;		/* current bucket */
};

static struct node *next_traverse_hash(struct traverse *where)
{
	if (where->b != (struct node *)NULL)
	{
		where->b = where->b->next;
	}
	while (where->b == (struct node *)NULL)
	{
		if (where->index >= where->h->num_buckets)
		{
			return((struct node *)NULL);
		}
		where->index++;
		where->b = where->h->all_buckets[where->index];
	}
	return(where->b);
}

static void start_traverse_hash(struct traverse *where, struct node *h)
{
	where->h = h;
	where->index = -1;
	where->b = (struct node *)NULL;
}

/* does not depend on which hash table is involved */
static struct node *make_string_node(char *key)
{
	struct node *temp;

	/* get the memory */
	temp = (struct node *) s_malloc(sizeof(struct node));
	uninit_assoc(temp);
	temp->hashvalue = hash_key(key);
	temp->keytype = STRING;
	temp->key.text = s_strdup(key);
	temp->type = NOTYPE;
	return(temp);
}

/* does not depend on which hash table is involved */
static struct node *make_number_node(int key)
{
	struct node *temp;

	/* get the memory */
	temp = (struct node *) s_malloc(sizeof(struct node));
	uninit_assoc(temp);
	temp->hashvalue = key;
	temp->keytype = NUMBER;
	temp->key.val = key;
	temp->type = NOTYPE;
	return(temp);
}

static void init_assoc(struct node *h, int buckets)
{
	int loop;

	/* init the important variables */
	h->count = 0;
	h->collisions = 0;
	h->num_buckets = buckets-1;	/* num_buckets is low by 1: & trick */

	/* sanity check */
	if (h->num_buckets < 0)
	{
		fprintf(stderr,"Must have at least one bucket for hashing.\n");
		exit(-1);
	}

	/* get the memory */
	h->all_buckets = (struct node **)
			s_malloc((h->num_buckets+1)*sizeof(struct node *));

	/* initialize to nothing */
	for(loop=0; loop<=h->num_buckets; loop++)
	{
		h->all_buckets[loop] = (struct node *)NULL;
	}
}

static void uninit_assoc(struct node *h)
{
	h->count = -1;		/* invalid flag */
	h->collisions = 0;
	h->num_buckets = -1;	/* invalid flag */
}

/* delete the bucket if it exists */
static void delete_node(struct node *it)
{
	unsigned int hashvalue;
	struct node *temp, **parent, *h;

	h = it->parent;
	hashvalue = it->hashvalue;
	parent = &(h->all_buckets[hashvalue&h->num_buckets]);
	temp = *parent;
	while (temp != (struct node *)NULL)
	{
		if (temp == it)
		{
			/* found it, so remove it. */
			*parent = temp->next;
			if (temp->next != (struct node *)NULL)
			{
				h->collisions--;
			}
			h->count--;
			cleanup_data(temp);
			if (temp->keytype == STRING)
			{
				free(temp->key.text);
			}
			cleanup_assoc(temp);
			free(temp);
			return;
		}
		parent = &(temp->next);
		temp = *parent;
	}
}

static void cleanup_assoc(struct node *h)
{
	int loop;
	struct node *temp;
	if (h->count < 0)
		return;		/* no data */

	for(loop=0; loop<=h->num_buckets; loop++)
	{
		temp = h->all_buckets[loop];
		while (temp != (struct node *)NULL)
		{
			h->all_buckets[loop] = h->all_buckets[loop]->next;
			cleanup_data(temp);
			cleanup_assoc(temp);	/* free the sub-hash */
			if (temp->keytype == STRING)
			{
				free(temp->key.text);
			}
			free(temp);
			temp = h->all_buckets[loop];
		}
	}
}

/* returns the appropriate bucket, creating it if it didn't exist */
static struct node *new_string_node(struct node *h, char *key)
{
	unsigned int hashvalue;
	struct node *temp;

	if (h->count < 0)
		init_assoc(h,1);
	hashvalue = hash_key(key);
	temp = h->all_buckets[hashvalue&h->num_buckets];
	while (temp != (struct node *)NULL)
	{
		if ((temp->keytype == STRING) &&
		    (hashvalue == temp->hashvalue) &&
		    !strcmp(key,temp->key.text))
		{
			/* found it */
			return(temp);
		}
		temp = temp->next;
	}
	if (h->all_buckets[hashvalue&h->num_buckets] != (struct node *)NULL)
	{
		if (h->collisions > h->num_buckets)
			resize_assoc(h, h->num_buckets * 2 + 2);
		else
			h->collisions++;
	}

	/* allocate a new bucket... */
	temp = make_string_node(key);

	h->count++;
	temp->next = h->all_buckets[hashvalue&h->num_buckets];
	h->all_buckets[hashvalue&h->num_buckets] = temp;
	temp->parent = h;
	return(temp);	/* which is NULL if we failed */
}

/* returns the appropriate bucket, creating it if it didn't exist */
static struct node *new_number_node(struct node *h, int key)
{
	unsigned int hashvalue;
	struct node *temp;

	if (h->count < 0)
		init_assoc(h,1);
	hashvalue = key;
	temp = h->all_buckets[hashvalue&h->num_buckets];
	while (temp != (struct node *)NULL)
	{
		if ((temp->keytype == NUMBER) &&
		    (hashvalue == temp->hashvalue) &&
		    (key == temp->key.val))
		{
			/* found it */
			return(temp);
		}
		temp = temp->next;
	}
	if (h->all_buckets[hashvalue&h->num_buckets] != (struct node *)NULL)
	{
		if (h->collisions > h->num_buckets)
			resize_assoc(h, h->num_buckets * 2 + 2);
		else
			h->collisions++;
	}

	/* allocate a new bucket... */
	temp = make_number_node(key);
	h->count++;
	temp->next = h->all_buckets[hashvalue&h->num_buckets];
	h->all_buckets[hashvalue&h->num_buckets] = temp;
	temp->parent = h;
	return(temp);
}

/* return the bucket only if it does exist */
static struct node *find_string_node(struct node *h, char *key)
{
	unsigned int hashvalue;
	struct node *temp;

	if (h->count < 0)
		return((struct node *)NULL);
	hashvalue = hash_key(key);
	temp = h->all_buckets[hashvalue&h->num_buckets];
	while (temp != (struct node *)NULL)
	{
		if ((temp->keytype == STRING) &&
		    (hashvalue == temp->hashvalue) &&
		    !strcmp(key,temp->key.text))
		{
			return(temp);
		}
		temp = temp->next;
	}
	return((struct node *)NULL);
}

/* return the bucket only if it does exist */
static struct node *find_number_node(struct node *h, int key)
{
	unsigned int hashvalue;
	struct node *temp;

	if (h->count < 0)
		return((struct node *)NULL);
	hashvalue = key;
	temp = h->all_buckets[hashvalue&h->num_buckets];
	while (temp != (struct node *)NULL)
	{
		if ((temp->keytype == NUMBER) &&
		    (hashvalue == temp->hashvalue) &&
		    (key == temp->key.val))
		{
			return(temp);
		}
		temp = temp->next;
	}
	return((struct node *)NULL);
}

static struct node *find_keyed_node(struct node *h, struct node *key)
{
	switch(key->type)
	{
		case STRING:
			return(find_string_node(h,key->u.text));
			break;
		case NUMBER:
			return(find_number_node(h,key->u.num));
			break;
		default:
			return((struct node *)NULL);
			break;
	}
}

static int resize_assoc(struct node *h, int buckets)
{
	struct node *temp, **old_buckets;
	int old_num_buckets, old_collisions, loop;

	/* sanity check */
	if (buckets < 1)
	{
		fprintf(stderr,"Must have at least one bucket for hashing.\n");
		return(-1);
	}
	/* check for power of two */
	if (buckets & (buckets-1))
	{
		fprintf(stderr,"Number of buckets must be a power of two.\n");
		return(-1);
	}

	old_num_buckets = h->num_buckets;
	old_buckets = h->all_buckets;
	old_collisions = h->collisions;

	h->collisions = 0;
	h->num_buckets = buckets-1;

	/* get the memory */
	h->all_buckets = (struct node **)
			s_malloc((h->num_buckets+1)*sizeof(struct node *));

	/* initialize to nothing -- hashes won't be the same */
	for(loop=0; loop<=h->num_buckets; loop++)
	{
		h->all_buckets[loop] = (struct node *)NULL;
	}

	/* now move stuff over */
	for (loop=0; loop<=old_num_buckets; loop++)
	{
		temp = old_buckets[loop];
		while (temp != (struct node *)NULL)
		{
			old_buckets[loop] = temp->next;
			
			if (h->all_buckets[temp->hashvalue&h->num_buckets] !=
					(struct node *)NULL)
				h->collisions++;
			temp->next = h->all_buckets[temp->hashvalue&h->num_buckets];
			h->all_buckets[temp->hashvalue&h->num_buckets] = temp;
			temp = old_buckets[loop];
		}
	}
	free(old_buckets);

	return(0);	/* success */
}

static void ntabs(FILE *fp, int level)
{
	for(;level;level--)
		putc('\t',fp);
}

static void print_escaped_string(FILE *fp, char *string)
{
	putc('\"',fp);
	while (*string)
	{
		switch(*string)
		{
			case '\a':	fputs("\\a",fp);	break;
			case '\b':	fputs("\\b",fp);	break;
			case '\t':	fputs("\\t",fp);	break;
			case '\n':	fputs("\\n",fp);	break;
			case '\v':	fputs("\\v",fp);	break;
			case '\f':	fputs("\\f",fp);	break;
			case '\r':	fputs("\\r",fp);	break;
			case '\"':	fputs("\\\"",fp);	break;
			case '\\':	fputs("\\\\",fp);	break;
			default:	putc(*string,fp);	break;
		}
		string++;
	}
	putc('\"',fp);
}

/* read from file, after starting ", until closing " */
static void read_escaped_string(FILE *fp, char *string, int maxlen)
{
	int len, c;

	len = 0;
	while (len < maxlen)
	{
		c = getc(fp);
		switch(c)
		{
			case EOF:
			case 0:
			case '\"':
				/* end of string */
				*string = 0;
				return;
				break;
			case '\\':
				c = getc(fp);
				switch(c)
				{
					case 'a':
						*string++ = '\a';
						break;
					case 'b':
						*string++ = '\b';
						break;
					case 't':
						*string++ = '\t';
						break;
					case 'n':
						*string++ = '\n';
						break;
					case 'v':
						*string++ = '\v';
						break;
					case 'f':
						*string++ = '\f';
						break;
					case 'r':
						*string++ = '\r';
						break;
					case '\"':
						*string++ = '\"';
						break;
					case '\\':
						*string++ = '\\';
						break;
					default:
						*string++ = 'c';
						break;
				}
				break;
			default:
				*string++ = c;
				break;
		}
		len++;
	}
	string--;
	*string = 0;
	return;
}

static void print_hash(struct node *h, FILE *fp, int level)
{
	struct node *temp;
	int loop;

	for(loop=0; loop<=h->num_buckets; loop++)
	{
		temp = h->all_buckets[loop];
		while (temp != (struct node *)NULL)
		{
			ntabs(fp,level);
			if (temp->keytype == STRING) {
				print_escaped_string(fp,temp->key.text);
			} else if (temp->keytype == NUMBER) {
				fprintf(fp,"%d",temp->key.val);
			} else {
				fprintf(fp,"Unknown");
			}
			fputs(" -> ",fp);
			print_data(fp, temp);
			putc('\n',fp);
			if (temp->count > 0)
			{
				ntabs(fp,level);
				fputs("{\n",fp);
				print_hash(temp,fp,level+1);
				ntabs(fp,level);
				fputs("}\n",fp);
			}
			temp = temp->next;
		}
	}
}

static void skip_arrow(FILE *fp)
{
	int c;

	if (!(((c = getc(fp)) == ' ') &&
	      ((c = getc(fp)) == '-') &&
	      ((c = getc(fp)) == '>') &&
	      ((c = getc(fp)) == ' ')))
	{
		fprintf(stderr, "Unexpected char '%c' in skip_arrow.\n",c);
		exit(-1);
	}
}

static void read_hash(struct node *h, FILE *fp)
{
	char rbuf[4096];
	int c, neg, intkey;
	struct node *temp;

	temp = (struct node *)NULL;
	neg = 0;
	while ((c = getc(fp)) != EOF)
	{
		switch (c)
		{
			case '\n':
			case '\t':
				/* ignore */
				break;
			case '{':
				if (temp != (struct node *)NULL)
				{
					read_hash(temp,fp);
				}
				break;
			case '}':
				return;
				break;
			case '\"':
				read_escaped_string(fp, rbuf, sizeof(rbuf));
				temp = new_string_node(h,rbuf);
				skip_arrow(fp);
				read_data(fp,temp);
				break;
			case '-':
				neg = 1;
				c = '0';
				/* and fall through */
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				intkey = 0;
				do {
					intkey *= 10;
					intkey += c - '0';
					c = getc(fp);
				}
				while (isdigit(c));
				if (neg)
				{
					intkey = -intkey;
					neg = 0;
				}
				ungetc(c,fp);
				temp = new_number_node(h,intkey);
				skip_arrow(fp);
				read_data(fp,temp);
				break;
			default:
				fprintf(stderr,
"Unexpected char '%c' in read_hash.\n",c);
				exit(-1);
				break;
		}
	}
}

static void log_event(char *where, char *message)
{
	char tbuffer[80];
	struct tm timestuff;
	time_t now;

	now = current_time.tv_sec;
	timestuff = *gmtime(&now);
	
	strftime(tbuffer,80,"(on %B %d %Y, at %H:%M:%S) ",&timestuff);
	if (where != (char *)NULL)
	{
		fprintf(stderr,"%s(in %s) %s",tbuffer,where,message);
	} else {
		fprintf(stderr,"%s%s",tbuffer,message);
	}
}

static void local_timestamp(int who)
{
	char tbuffer[80];
	struct tm timestuff, modtime;
	time_t now;

	now = current_time.tv_sec;
	timestuff = *gmtime(&now);

	modtime = timestuff;
	modtime.tm_hour += p[who].timezone;
	while (modtime.tm_hour < 0)
		modtime.tm_hour += 24;
	while (modtime.tm_hour > 23)
		modtime.tm_hour -= 24;
	strftime(tbuffer,80,"(at %H:%M:%S) ",&modtime);
	send_string(who,tbuffer,0,-1);
}

static void send_string(int who, char *str, int start, int class)
{
	int wcount;
	char outbuf[OUT_BUFFER_SIZE*3];
	char *from, *to;

	/* blank line to clear partial input line */
	if (start && (class != CLASS_SELF))
	{
		p[who].message(p[who].id,"\r\n",2,0);
	}
	if (start && (p[who].timestamp & class))
	{
		local_timestamp(who);
	}
	if (p[who].wrap > 0)
	{
		to = outbuf;
		from = str;
		wcount = 0;
		while (*from)
		{
			if ((*to++ = *from++) == '\n')
			{
				wcount = 0;
			} else {
				wcount++;
			}
			if (wcount == p[who].wrap)
			{
				*to++ = '\r';
				*to++ = '\n';
				wcount = 0;
			}
		}
		*to = 0;
		p[who].message(p[who].id,outbuf,-1,0);
	} else {
		p[who].message(p[who].id,str,-1,0);
	}
}

static int get_named_number(struct node *ob, char *name)
{
	struct node *temp;

	temp = find_string_node(ob,name);
	if ((temp != (struct node *)NULL) &&
	    (temp->type == NUMBER))
	{
		return(temp->u.num);
	} else {
		return(0);
	}
}

static char *get_named_string(struct node *ob, char *name)
{
	struct node *temp;

	temp = find_string_node(ob,name);
	if ((temp != (struct node *)NULL) &&
	    (temp->type == STRING))
	{
		return(temp->u.text);
	} else {
		return("");
	}
}

static struct node *get_named_property(struct node *ob, char *name)
{
	return(new_string_node(ob,name));
}

static void clear_named_property(struct node *ob, char *name)
{
	struct node *temp;

	temp = new_string_node(ob,name);
	if (temp->type == STRING)
	{
		free(temp->u.text);
	}
	temp->type = NOTYPE;
}

static void set_named_number(struct node *ob, char *name, int val)
{
	struct node *temp;

	temp = new_string_node(ob,name);
	if (temp->type == STRING)
	{
		free(temp->u.text);
	}
	temp->type = NUMBER;
	temp->u.num = val;
}

static void set_named_string(struct node *ob, char *name, char *val)
{
	struct node *temp;

	temp = new_string_node(ob,name);
	if (temp->type == STRING)
	{
		free(temp->u.text);
	}
	temp->type = STRING;
	temp->u.text = s_strdup(val);
}

static void set_named_property(struct node *ob, char *name, struct node *val)
{
	struct node *temp;

	temp = new_string_node(ob,name);
	if (temp->type == STRING)
	{
		free(temp->u.text);
	}
	if (val != (struct node *)NULL)
	{
		temp->type = val->type;
		switch(val->type)
		{
			case NOTYPE:
				/* done */
				break;
			case STRING:
				temp->u.text = s_strdup(val->u.text);
				break;
			case NUMBER:
				temp->u.num = val->u.num;
				break;
			default:
				fprintf(stderr,"Copying unknown type!\n"); /* oops */
				break;
		}
	} else {
		temp->type = NOTYPE;
	}
}

static void tell_object(struct node *mob, char *str, int class)
{
	int who;

	who = get_named_number(mob,"player");
	if (who != -1)
	{
		send_string(who, str, 1, class);
	}
}

static void player_broadcast(char *str, int start, int class)
{
	int loop;

	for (loop=0;loop<max_players;loop++)
	{
		if (p[loop].state != -1)
		{
			send_string(loop,str,start,class);
		}
	}
	/* For logging of the local discussion */
	log_event(NULL,str);
}

static struct node *find_where(struct node *ob)
{
	struct node *where;

	where = find_string_node(ob,"where");
	if (where != (struct node *)NULL)
	{
		return(find_keyed_node(objects,where));
	} else {
		return((struct node *)NULL);
	}
}

static struct node *find_firstinv(struct node *ob)
{
	struct node *inv;

	inv = find_string_node(ob,"inv");
	if (inv != (struct node *)NULL)
	{
		return(find_keyed_node(objects,inv));
	} else {
		return((struct node *)NULL);
	}
}

static struct node *find_nextinv(struct node *ob)
{
	struct node *nextinv;

	nextinv = find_string_node(ob,"nextinv");
	if (nextinv != (struct node *)NULL)
	{
		return(find_keyed_node(objects,nextinv));
	} else {
		return((struct node *)NULL);
	}
}

static void tell_room(struct node *ob, char *str, struct node *ex1, struct node *ex2, int class)
{
	struct node *t_ob;

	t_ob = find_firstinv(ob);
	while (t_ob != (struct node *)NULL)
	{
		if ((t_ob != ex1) &&
		    (t_ob != ex2))
		{
			tell_object(t_ob,str,class);
		}
		t_ob = find_nextinv(t_ob);
	}
}

static void tell_other_rooms(struct node *root, struct node *ex, char *str, int class)
{
	struct node *t_ob, *plbucket;
	int who;

	t_ob = find_firstinv(root);
	while (t_ob != (struct node *)NULL)
	{
		if (t_ob != ex)
		{
			plbucket = find_string_node(t_ob,"player");
			if (plbucket != (struct node *)NULL)
			{
				who = plbucket->u.num;
				if (who != -1)
				{
					if (p[who].timestamp & class)
					{
						local_timestamp(who);
					}
					send_string(who, str, 1, class);
				}
			}
			tell_other_rooms(t_ob, ex, str, class);
		}
		t_ob = find_nextinv(t_ob);
	}
}

static void broadcast(char *str, int start, int class)
{
	tell_other_rooms(root_object, (struct node *)NULL, str, class);
}

static struct node *new_object(char *name)
{
	struct node *this;
	int count;

	if (name == (char *)NULL)
	{
		count = obcount->u.num++;
		this = new_number_node(objects,count);
		set_named_number(this,"key",count);
	} else {
		this = new_string_node(objects,name);
		set_named_string(this,"key",name);
	}

	/* left default/unset: name, where, inv, nextinv, exits */
	set_named_number(this,"player",-1);	/* no connected player */
	set_named_number(this,"flags",O_NONE);	/* no flags */
	clear_named_property(this,"inv");
	clear_named_property(this,"nextinv");
	clear_named_property(this,"where");
	clear_named_property(this,"name");
	clear_named_property(this,"exits");
	return(this);
}

/* don't call if it isn't in a location... */
static void remove_object(struct node *it)
{
	struct node *ob, *where, *nextinv;

	where = find_where(it);
	ob = find_firstinv(where);
	if (ob == it)
	{
		set_named_property(where,"inv",find_string_node(it,"nextinv"));
	} else {
		/* note: comma operator */
		while (nextinv = find_nextinv(ob), nextinv != it)
		{
			ob = nextinv;
		}
		set_named_property(ob,"nextinv",find_string_node(it,"nextinv"));
	}
	set_named_number(it,"where",0); /* nowhere */
	set_named_number(it,"nextinv",0); /* no next object nowhere */
}

/* requires that no exits currently point to it */
static void free_object(struct node *it)
{
	if (find_string_node(it,"where") != (struct node *)NULL)
	{
		remove_object(it);
	}
	/* remove inventory */
	while (find_string_node(it,"inv")->type != NOTYPE)
	{
		free_object(find_string_node(it,"inv"));
	}
	/* name automatically freed */
	/* exits automatically freed */
	delete_node(it);
}

/* used in assoc stuff */
void cleanup_data(struct node *it)
{
	switch (it->type)
	{
		case NOTYPE:
			/* no cleanup required */
			break;
		case STRING:
			if (it->u.text != (char *)NULL)
			{
				free(it->u.text);
			}
			break;
		case NUMBER:
			/* no cleanup required */
			break;
		default:
			fprintf(stderr,"Fatal: unknown data type in hash.\n");
			exit(-1);
	}
}

void print_data(FILE *fp, struct node *it)
{
	switch (it->type)
	{
		case NOTYPE:
			fprintf(fp,"Null entry");
			break;
		case STRING:
			print_escaped_string(fp,it->u.text);
			break;
		case NUMBER:
			fprintf(fp,"%d", it->u.num);
			break;
		default:
			fprintf(stderr,"Fatal: unknown data type in hash.\n");
			exit(-1);
	}
}

void read_data(FILE *fp, struct node *it)
{
	int c, neg, intkey;
	char buffer[4096];

	c = getc(fp);
	neg = 0;
	switch (c)
	{
		case 'N':
			fgets(buffer,10,fp);
			if (strcmp(buffer,"ull entry"))
			{
				fprintf(stderr,"Fatal: unknown data type in read_data.\n");
			}
			it->type = NOTYPE;
			break;
		case '"':
			it->type = STRING;
			read_escaped_string(fp,buffer,sizeof(buffer));
			it->u.text = s_strdup(buffer);
			break;
		case '-':
			neg = 1;
			c = '0';
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			intkey = 0;
			do {
				intkey *= 10;
				intkey += c - '0';
				c = getc(fp);
			}
			while (isdigit(c));
			if (neg)
			{
				intkey = -intkey;
				neg = 0;
			}
			it->type = NUMBER;
			it->u.num = intkey;
			break;
		case EOF:
			fprintf(stderr,"Fatal: EOF in read_data.\n");
			break;
		default:
			fprintf(stderr,"Fatal: unknown data type in read_data.\n");
			break;
	}
}

/* check for loop? */
static void insert_object(struct node *it, char *where)
{
	struct node *obwhere;

	if (find_where(it) != (struct node *)NULL)
	{
		remove_object(it);
	}
	set_named_string(it,"where",where);
	obwhere = find_string_node(objects,where);
	set_named_property(it,"nextinv",get_named_property(obwhere,"inv"));
	set_named_property(obwhere,"inv",get_named_property(it,"key"));
}

static void new_room(char *name, char *shortdesc)
{
	struct node *spot;

	/* exist already -- overwrite */
	spot = new_string_node(objects,name);
	set_named_string(spot,"name",name);
	set_named_string(spot,"key",name);
	set_named_string(spot,"shortdesc",shortdesc);
	insert_object(spot, "root object");
}

/* TODO: expand to deal with numbered rooms? */
static struct node *named_object(char *objname)
{
	struct node *obj;

	obj = find_string_node(objects,objname);
	return(obj);
}

static void link_exit(char *dir, char *here, char *there)
{
	struct node *here_room, *exits;

	here_room = named_object(here);
	if (here_room == (struct node *)NULL)
	{
		fprintf(stderr,"linking from nonexistent room %s\n",here);
		return;
	}
	exits = new_string_node(here_room,"exits");
	set_named_string(exits,dir,there);
}

static int new_player(int id,void (*message)(int id,char *message,int len,int oob))
{
	char output_buffer[OUT_BUFFER_SIZE];
	struct node *playerob, *tempbucket;
	int who;

	/* find a player id */
	for (who=0;
	     (who<max_players) && (p[who].state != -1);
	     who++)
		;	/* do nothing */
	if (who >= max_players)
	{
		/* TODO bail on player, or expand sockets, (realloc p) or... */
	}
	
	playerob = new_object((char *)NULL);

	p[who].state = 0;
	p[who].wrap = 0;
	p[who].command_tail = 0;
	p[who].object = playerob;
	p[who].message = message;
	p[who].id = id;
	p[who].timezone = 0;
	p[who].timestamp = 0;
	tempbucket = new_string_node(playerob,"player");
	tempbucket->type = NUMBER;
	tempbucket->u.num = who;
	sprintf(output_buffer,"Player %d", who);
	set_named_string(playerob,"name",output_buffer);
	insert_object(playerob, "start room");

	sprintf(output_buffer,"Welcome, %d! (%s)\r\n",
		who, connection_name(p[who].id));
	broadcast(output_buffer,1,CLASS_SYSTEM);
	return(who);	/* what ID to use to contact us */
}

static void do_tell(int who, char *command, char *arg);
static void do_kill(int who, char *command, char *arg);
static void do_dump(int who, char *command, char *arg);
static void do_loadauth(int who, char *command, char *arg);
static void do_broadcast(int who, char *command, char *arg);
static void do_user(int who, char *command, char *arg);
static void do_pass(int who, char *command, char *arg);
static void do_testcommand(int who, char *command, char *arg);
static void do_startchat(int who, char *command, char *arg);
static void do_startcommandchat(int who, char *command, char *arg);
static void do_startpaste(int who, char *command, char *arg);
static void do_endchat(int who, char *command, char *arg);
static void do_speech(int who, char *command, char *arg);
static void do_emote(int who, char *command, char *arg);
static void do_nospace_emote(int who, char *command, char *arg);
static void do_help(int who, char *command, char *arg);
static void do_name(int who, char *command, char *arg);
static void do_who(int who, char *command, char *arg);
static void do_wrap(int who, char *command, char *arg);
static void do_quit(int who, char *command, char *arg);
static void do_where(int who, char *command, char *arg);
static void do_look(int who, char *command, char *arg);
static void do_zone(int who, char *command, char *arg);
static void do_timestamp(int who, char *command, char *arg);
static void do_go(int who, char *command, char *arg);
static void do_namedir(int who, char *command, char *arg);

struct commandinfo {
	char *name;
	int length, flags;
	void (*func)(int who, char *command, char *arg);
	char *helpstring, *longhelpstring;
} commands[] = {
	{ "chat", 4, O_NONE, do_startchat,
		"\"chat\": say strings",
		"chat: enter a mode where text typed isn't interpreted as "
		"commands, but rather said.  Ended by a single '.' alone on "
		"a line." },
	{ "commandchat", 11, O_NONE, do_startcommandchat,
		"\"commandchat\": say strings that aren't commands",
		"commandchat: enter a mode where text typed is interpreted as "
		"commands, but if no command matches, is instead spoken.  "
		"Ended by a single '.' alone on a line." },
	{ "paste", 11, O_NONE, do_startpaste,
		"\"paste\": say a block of text at once",
		"paste: say a long utterance, without having to prefix "
		" each line with a command, and rendering the output as "
		" a single statement.  Ended with '.' alone on a line." },
	{ ".", 1, O_NONE, do_endchat,
		"\".\": end a chat mode",
		".: End a chat mode:  Return to the default command mode."},
	{ "\"", 1, O_ARG, do_speech,
		"\"\' or \"\": say a string",
		"\":  Say something -- the rest of the line.  If the line "
		" ends with a '?' or '!', it will be asked or exclaimed, "
		" respectively."},
	{ "'", 1, O_ARG | O_HIDDEN, do_speech,
		"\"'\": say a string",
		"\":  Say something -- the rest of the line.  If the line "
		" ends with a '?' or '!', it will be asked or exclaimed, "
		" respectively."},
	{ "think ", 6, O_ARG, do_speech,
		"\"% or think \": think a string",
		"think:  Make a thought bubble, with the typed text inside."},
	{ "%", 1, O_ARG | O_HIDDEN, do_speech,
		"\"%\": think a string",
		"%: (think) "
		"Make a thought bubble, with the typed text inside."},
	{ "say ", 4, O_ARG, do_speech,
		"\"say, mumble, exclaim, ask, sing \": say a string",
		"say:  Say something -- the rest of the line.  If the line "
		" ends with a '?' or '!', it will be asked or exclaimed, "
		" respectively."},
	{ "mumble ", 7, O_ARG | O_HIDDEN, do_speech,
		"\"mumble \": mumble a string",
		"mumble:  Mumble something -- the rest of the line."},
	{ "sing ", 5, O_ARG | O_HIDDEN, do_speech,
		"\"sing \": sing <text to be sung>",
		"sing:  Sing something -- the rest of the line."},
	{ "exclaim ", 8, O_ARG | O_HIDDEN, do_speech,
		"\"exclaim \": exclaim <text to be exclaimed>",
		"exclaim:  This is a specific version of the say commands. "
		"The specified text will be exclaimed."},
	{ "ask ", 4, O_ARG | O_HIDDEN, do_speech,
		"\"ask \": ask <text to be asked>",
		"ask:  This is a specific version of the say commands. "
		"The specified text will be asked."},
	{ "emote ", 6, O_ARG, do_emote,
		"\"emote \": emote a string",
		"emote:  Display the typed text as an action taken."},
	{ ":", 1, O_ARG, do_emote,
		"\":\": emote a string",
		":: (emote) Display the typed text as an action taken."},
	{ ";", 1, O_ARG, do_nospace_emote,
		"\";\": emote a string with no extra space",
		";: (emote) Display the typed text as an action taken, with "
		"no space after the player name (useful for posessives, etc."},
	{ "help ", 5, O_ARG | O_HIDDEN, do_help,
		"\"help\": print commands",
		"help:  Print the list of commands, or help for a specific "
		"command."},
	{ "help", 4, O_NONE, do_help,
		"\"help\": print commands",
		"help:  Print the list of commands, or help for a specific "
		"command."},
	{ "name ",5, O_ARG, do_name,
		"\"name \": set your name",
		"name:  Set or erase your player name.  The player name "
		"is displayed in parenthesis for says, emotes, etc."},
	{ "name",4, O_NONE, do_name,
		"\"name \": set your name",
		"name:  Set or erase your player name.  The player name "
		"is displayed in parenthesis for says, emotes, etc."},
	{ "who", 3, O_NONE, do_who,
		"\"who\": list connected players",
		"who:  List all connected players by name and number."},
	{ "w", 1, O_NONE, do_who,
		"\"w\": alias for who",
		"who:  List all connected players by name and number."},
	{ "wrap ", 5, O_ARG, do_wrap,
		"\"wrap \": set your connection to wrap at n characters",
		"wrap:  Set line-wrapping at <argument> characters.  With no "
		"argument, turn off line-wrapping."},
	{ "wrap", 4, O_NONE, do_wrap,
		"\"wrap\" (with no argument): set your connection to not wrap",
		"wrap:  Set line-wrapping at <argument> characters.  With no "
		"argument, turn off line-wrapping."},
	{ "look", 4, O_NONE, do_look,
		"\"look\": examine local environment",
		"look:  Get a description of the current room, along with "
		"the people and things in it."},
	{ "l", 1, O_NONE, do_look,
		"\"l\": alias for look",
		"l:  (short for look) Get a description of the current room, "
		"along with the people and things in it."},
	{ "zone ", 5, O_ARG, do_zone,
		"\"zone \": set your timezone",
		"zone:  Set the timezone, so that timestamps (if enabled) "
		"show up in your preferred timezone.  The argument should "
		"be a number of hours (possibly negative) to add to UTC."},
	{ "timezone ", 9, O_ARG, do_zone,
		"\"timezone \": set your timezone",
		"timezone:  Set the timezone, so that timestamps (if enabled) "
		"show up in your preferred timezone.  The argument should "
		"be a number of hours (possibly negative) to add to UTC."},
	{ "timestamp ", 10, O_ARG, do_timestamp,
		"\"timestamp \": set to timestamp no, some, or all events",
		"timestamp:  Tell the game to turn off timestamps (timestamp "
		"no), timestamp for only some events, like player connects "
		" and disconnects (timestamp some), or timestamp everything "
		"(timestamp all).  Arguments of system, mob, other or self "
		" toggle just those classes of events This interacts with "
		"the (time)zone command."},
	{ "quit", 4, O_NONE, do_quit,
		"\"quit\": leave the game",
		"quit:  Disconnect from the game."},
	{ "where", 5, O_NONE, do_where,
		"\"where\": tell what room you're in",
		"where:  State which room you're in."},
	{ "go ", 3, O_ARG, do_go,
		"\"go \": travel in some direction",
		"go:  Travel in the specified direction, if possible."},
	{ "up", 5, O_NONE, do_namedir,
		"\"up, down, north, south, east, west ...\": travel in that direction",
		"up:  Travel up, if possible."},
	{ "down", 5, O_HIDDEN, do_namedir,
		"\"down\": travel down",
		"down:  Travel down, if possible."},
	{ "north", 5, O_HIDDEN, do_namedir,
		"\"north\": travel north",
		"north:  Travel north, if possible."},
	{ "south", 5, O_HIDDEN, do_namedir,
		"\"south\": travel south",
		"south:  Travel south, if possible."},
	{ "east", 5, O_HIDDEN, do_namedir,
		"\"east\": travel east",
		"east:  Travel east, if possible."},
	{ "west", 5, O_HIDDEN, do_namedir,
		"\"west\": travel west",
		"west:  Travel west, if possible."},
	{ "northeast", 5, O_HIDDEN, do_namedir,
		"\"northeast\": travel northeast",
		"northeast:  Travel northeast, if possible."},
	{ "northwest", 5, O_HIDDEN, do_namedir,
		"\"northwest\": travel northwest",
		"northwest:  Travel northwest, if possible."},
	{ "southeast", 5, O_HIDDEN, do_namedir,
		"\"southeast\": travel southeast",
		"southeast:  Travel southeast, if possible."},
	{ "southwest", 5, O_HIDDEN, do_namedir,
		"\"southwest\": travel southwest",
		"southwest:  Travel southwest, if possible."},
	{ "test ", 5, O_ARG, do_testcommand,
		"\"test\": a test command",
		"test:  A useless (vestigal) test command."},
	{ "kill ", 5, O_ARG, do_kill,
		"\"kill \": try to kill something",
		"kill:  Try to kill something (not very effectively)."},
	{ "tell ", 5, O_ARG, do_tell,
		"\"tell \": tell player (identified by number) something",
		"tell:  Tell another player a message (identified by player "
		"number)."},
	{ "dump", 4, O_PRIV | O_HIDDEN, do_dump,
		"\"dump\": dump the database",
		"dump:  Dump the database to log (debug)."},
	{ "loadauth", 4, O_PRIV | O_HIDDEN, do_loadauth,
		"\"loadauth\": load the auth database",
		"loadauth:  Load the auth database."},
	{ "broadcast ", 10, O_PRIV | O_HIDDEN | O_ARG, do_broadcast,
		"\"broadcast\": echo a message to everyone",
		"broadcast:  echo a message to everyone."},
	{ "user ", 5, O_ARG, do_user,
		"\"user \": start login",
		"user:  Specify a username to log in as."},
	{ "pass ", 5, O_ARG, do_pass,
		"\"pass \": finish login",
		"pass:  Specify a password for login."},
	{ NULL, 0, O_INVALID, NULL,
		"\"\": no command",
		"empty command placeholder."},
};

static void list_who(int who)
{
	char output_buffer[OUT_BUFFER_SIZE];
	int loop, secdelta, usecdelta;
	struct timeval idletime;
	char *timename;

	send_string(who,"*: Player -1 (logger)\r\n",1,CLASS_SELF);
	for (loop=0;loop<max_players;loop++)
	{
		if (p[loop].state != -1)
		{
			timename = "second";
			idletime = connection_time(p[loop].id);
			usecdelta = current_time.tv_usec - idletime.tv_usec;
			secdelta = current_time.tv_sec - idletime.tv_sec;
			sprintf(output_buffer,"%d: ",loop);
			send_string(who,output_buffer,0,CLASS_SELF);
			send_string(who,
				get_named_string(p[loop].object,"name")
					,0,CLASS_SELF);
			if (usecdelta < 0)
			{
				secdelta--;
				usecdelta += 1000000;
			}
			if (secdelta >= 60)
			{
				secdelta = secdelta/60;
				timename = "minute";
				if (secdelta >= 60)
				{
					secdelta = secdelta/60;
					timename = "hour";
					if (secdelta >= 24)
					{
						secdelta = secdelta/24;
						timename = "day";
						if (secdelta >= 7)
						{
							secdelta = secdelta/7;
							timename = "week";
						}
					}
				}
			}
			if (secdelta != 0)
			{
				sprintf(output_buffer,
					" idle %d %s", secdelta, timename);
				send_string(who,output_buffer,0,CLASS_SELF);
				if (secdelta != 1)
					send_string(who,"s",0,CLASS_SELF);
			} else if (usecdelta != 0) {
				sprintf(output_buffer,
					" idle %d microsecond", usecdelta);
				send_string(who,output_buffer,0,CLASS_SELF);
				if (usecdelta != 1)
					send_string(who,"s",0,CLASS_SELF);
			}
			/* else not idle */
			send_string(who,"\r\n",0,CLASS_SELF);
		}
	}
}

static void do_testcommand(int who, char *command, char *arg)
{
	send_string(who, "Test command: \"", 1, CLASS_SELF);
	send_string(who, arg, 1, CLASS_SELF);
	send_string(who, "\"\r\n" , 1, CLASS_SELF);
}

static void do_startchat(int who, char *command, char *arg)
{
	p[who].state = 1;	/* player is in chat mode */
	send_string(who,
		"Now in chat mode.  End with \".\" alone on a line.\r\n",
			1, CLASS_SELF);
}

static void do_startcommandchat(int who, char *command, char *arg)
{
	if (p[who].state == 2)
	{
		send_string(who,
"Still in commandchat mode.  End with \".\" alone on a line.\r\n",
			1, CLASS_SELF);
	} else {
		p[who].state = 2;	/* player is in commandchat mode */
		send_string(who,
 "Now in commandchat mode.  End with \".\" alone on a line.\r\n",
			1, CLASS_SELF);
	}
}

static void do_startpaste(int who, char *command, char *arg)
{
	p[who].state += 3;	/* player is in a paste mode */
	p[who].paste_buffer[0] = 0;
	send_string(who,
		"Now in paste mode.  End with \".\" alone on a line.\r\n",
			1, CLASS_SELF);
}

struct speech {
	char *command;
	char *prefix;
	char *suffix;
} speechtypes[] = {
	{ "say", " says: ", "" },
	{ "ask", " asks: ", "" },
	{ "exclaim", " exclaims: ", "" },
	{ "mumble", " mumbles: ", "" },
	{ "think", " . o O ( ", " )" },
	{ "%", " . o O ( ", " )" },
	{ "sing", " sings o/~ ", " o/~ " },
};


static void do_speech(int who, char *command, char *arg)
{
	char output_buffer[OUT_BUFFER_SIZE];
	int len, i;
	char *prefix, *suffix;

	suffix = "";
	if ((command[0] == '\'') ||
	    (command[0] == '\"') ||
	    (command[0] == 0))
	{
		len = strlen(arg);
		if (arg[len-1] == '!')
		{
			prefix = " exclaims: ";
		} else if (arg[len-1] == '?')
		{
			prefix = " asks: ";
		} else {
			prefix = " says: ";
		}
	} else {
		prefix = " says: ";
		for (i=0; i<(sizeof(speechtypes)/sizeof(speechtypes[0])); i++)
		{
			if (!strncmp(command,
				     speechtypes[i].command,
				     strlen(speechtypes[i].command)))
			{
				prefix = speechtypes[i].prefix;
				suffix = speechtypes[i].suffix;
				break;
			}
		}
	}

	sprintf(output_buffer,"%s%s%s%s\r\n",
		get_named_string(p[who].object,"name"), prefix, arg, suffix);
	broadcast(output_buffer,1,CLASS_OTHER);
}

static void do_emote(int who, char *command, char *arg)
{
	char output_buffer[OUT_BUFFER_SIZE];

	sprintf(output_buffer,"%s %s\r\n",
		get_named_string(p[who].object,"name"), arg);
	broadcast(output_buffer,1,CLASS_OTHER);
}

static void do_nospace_emote(int who, char *command, char *arg)
{
	char output_buffer[OUT_BUFFER_SIZE];

	sprintf(output_buffer,"%s%s\r\n",
		get_named_string(p[who].object,"name"), arg);
	broadcast(output_buffer,1,CLASS_OTHER);
}

static void do_who(int who, char *command, char *arg)
{
	list_who(who);
}

static void do_help(int who, char *command, char *arg)
{
	int index, len;

	if (*arg)
	{
		index = 0;
		len = strlen(arg);
		while (commands[index].flags != O_INVALID)
		{
			if ((!(strncmp(commands[index].name,arg,len))) &&
			    ((commands[index].name[len] == '\0') ||
			     (commands[index].name[len] == ' ')))
			{
				send_string(who,
					commands[index].longhelpstring,
						1, CLASS_SELF);
				send_string(who, "\r\n", 1, CLASS_SELF);
				return;
			}
			index++;
		}
		send_string(who, "No command \"", 1, CLASS_SELF);
		send_string(who, arg, 1, CLASS_SELF);
		send_string(who, "\" exists.\r\n", 1, CLASS_SELF);
	} else {
		send_string(who, "Commands are:\r\n", 1, CLASS_SELF);
		index = 0;
		while (commands[index].flags != O_INVALID)
		{
			if (!(commands[index].flags & O_HIDDEN))
			{
				send_string(who, commands[index].helpstring,
					0, CLASS_SELF);
				send_string(who, ".\r\n", 0, CLASS_SELF);
			}
			index++;
		}
		send_string(who, "Big world version number is 16.\r\n", 0, CLASS_SELF);
	}
}

static void do_wrap(int who, char *command, char *arg)
{
	char output_buffer[OUT_BUFFER_SIZE];
	int wraptemp;

	wraptemp = atoi(arg);
	if (wraptemp > 0)
	{
		sprintf(output_buffer,
			"Wrapping at %d characters.\r\n", wraptemp);
		send_string(who, output_buffer, 0, CLASS_SELF);
		p[who].wrap = wraptemp;
	} else {
		send_string(who, "Wrapping disabled.\r\n", 0, CLASS_SELF);
		p[who].wrap = 0;
	}
}

static void do_name(int who, char *command, char *arg)
{
	char output_buffer[OUT_BUFFER_SIZE];

	if (strlen(arg) < 85)
	{
		if (arg[0])
		{
			sprintf(output_buffer,"Player %d (%s)", who, arg);
		} else {
			sprintf(output_buffer,"Player %d", who);
		}
		set_named_string(p[who].object,"name",output_buffer);
		send_string(who, "Now ", 0, CLASS_SELF);
		send_string(who, get_named_string(p[who].object,"name"),
			0, CLASS_SELF);
		send_string(who, "\r\n", 0, CLASS_SELF);
	}
}

static void do_endchat(int who, char *command, char *arg)
{
	if (p[who].state == 0)
	{
		send_string(who, "Ending nonexistent chat mode.\r\n",
			0, CLASS_SELF);
	} else {
		send_string(who, "Ending chat mode.\r\n", 0, CLASS_SELF);
	}
	p[who].state = 0;
}

static void do_quit(int who, char *command, char *arg)
{
	char output_buffer[OUT_BUFFER_SIZE];

	sprintf(output_buffer,"%s quit.\r\n",get_named_string(p[who].object,"name"));
	broadcast(output_buffer,1,CLASS_SYSTEM);
	if (p[who].object != (struct node *)NULL)
		free_object(p[who].object);
	p[who].object = (struct node *)NULL;
	close_connection(p[who].id);
}

static void do_where(int who, char *command, char *arg)
{
	struct node *ob;

	ob = p[who].object;
	send_string(who, "In ", 0, CLASS_SELF);
	send_string(who, get_named_string(ob,"where"), 0, CLASS_SELF);
	send_string(who, ".\r\n", 0, CLASS_SELF);
}

static void do_look(int who, char *command, char *arg)
{
	struct node *ob, *room, *others, *exits, *tempbucket;
	struct traverse t;

	ob = p[who].object;
	room = find_where(ob);
	send_string(who, get_named_string(room,"shortdesc"), 1, CLASS_SELF);
	send_string(who, ".\r\n", 0, CLASS_SELF);
	exits = find_string_node(room,"exits");
	if (exits != (struct node *)NULL)
	{
		send_string(who, "Exits are:", 0, CLASS_SELF);
		start_traverse_hash(&t,exits);
		while ((tempbucket = next_traverse_hash(&t)) !=
		       (struct node *)NULL)
		{
			send_string(who, " ", 0, CLASS_SELF);
			send_string(who, tempbucket->key.text, 0, CLASS_SELF);
		}
		send_string(who, "\r\n", 0, CLASS_SELF);
	}
	others = find_firstinv(room);
	while (others != (struct node *)NULL)
	{
		if ((ob != others) &&
		    !(get_named_number(others,"flags") & INVIS_FLAG))
		{
			send_string(who, get_named_string(others,"name"),
				0, CLASS_SELF);
			send_string(who, " is here.\r\n", 0, CLASS_SELF);
		}
		others = find_nextinv(others);
	}
}

static void do_zone(int who, char *command, char *arg)
{
	char output_buffer[OUT_BUFFER_SIZE];
	int newzone;	/* long, but I want a -12 -> +12 range or so */

	char *endptr;

	newzone = strtol(arg, &endptr, 10);
	if ((*endptr != 0) || (endptr == arg))
	{
		send_string(who, "Can't convert \"", 0, CLASS_SELF);
		send_string(who, arg, 0, CLASS_SELF);
		send_string(who, "\" into a time zone number.\r\n",
			0, CLASS_SELF);
	} else {
		p[who].timezone = newzone;
		sprintf(output_buffer,"New time zone offset is %d.\r\n", newzone);
		send_string(who, output_buffer, 0, CLASS_SELF);
	}
}

static void do_timestamp(int who, char *command, char *arg)
{
	if (!strcmp(arg,"no") || !strcmp(arg,"none"))
	{
		p[who].timestamp = 0;
		send_string(who, "Now timestamping no events.\r\n",
			0, CLASS_SELF);
	}
	else if (!strcmp(arg,"some"))
	{
		p[who].timestamp = CLASS_SYSTEM;
		send_string(who, "Now timestamping some events.\r\n",
			0, CLASS_SELF);
	}
	else if (!strcmp(arg,"all"))
	{
		p[who].timestamp = CLASS_SYSTEM | CLASS_MOB | CLASS_OTHER;
		send_string(who, "Now timestamping all events.\r\n",
			0, CLASS_SELF);
	}
	else if (!strcmp(arg,"system"))
	{
		p[who].timestamp ^= CLASS_SYSTEM;
		send_string(who, "Timestamping for system events turned ",
			0, CLASS_SELF);
		if (p[who].timestamp & CLASS_SYSTEM)
		{
			send_string(who, "on.\r\n", 0, CLASS_SELF);
		} else {
			send_string(who, "off.\r\n", 0, CLASS_SELF);
		}
	}
	else if (!strcmp(arg,"mob"))
	{
		p[who].timestamp ^= CLASS_MOB;
		send_string(who, "Timestamping for mob events turned ",
			0, CLASS_SELF);
		if (p[who].timestamp & CLASS_MOB)
		{
			send_string(who, "on.\r\n", 0, CLASS_SELF);
		} else {
			send_string(who, "off.\r\n", 0, CLASS_SELF);
		}
	}
	else if (!strcmp(arg,"other"))
	{
		p[who].timestamp ^= CLASS_OTHER;
		send_string(who, "Timestamping for other events turned ",
			0, CLASS_SELF);
		if (p[who].timestamp & CLASS_OTHER)
		{
			send_string(who, "on.\r\n", 0, CLASS_SELF);
		} else {
			send_string(who, "off.\r\n", 0, CLASS_SELF);
		}
	}
	else if (!strcmp(arg,"self"))
	{
		p[who].timestamp ^= CLASS_SELF;
		send_string(who, "Timestamping for self events turned ",
			0, CLASS_SELF);
		if (p[who].timestamp & CLASS_SELF)
		{
			send_string(who, "on.\r\n", 0, CLASS_SELF);
		} else {
			send_string(who, "off.\r\n", 0, CLASS_SELF);
		}
	} else {
		send_string(who,"Can't convert \"",0,CLASS_SELF);
		send_string(who,arg,0,CLASS_SELF);
		send_string(who,"\" into a timestamp type.\r\n",0,CLASS_SELF);
	}
}

static void mob_look(struct node *mob)
{
	if (get_named_number(mob,"player") != -1)
	{
		do_look(get_named_number(mob,"player"), "", "");
	}
}

static char *random_exit(struct node *room)
{
	int exitcount, exitchoice;
	struct node *tempbucket, *exits;
	struct traverse t;

	exitcount = 0;
	exits = find_string_node(room,"exits");
	start_traverse_hash(&t,exits);
	while ((tempbucket = next_traverse_hash(&t)) !=
	       (struct node *)NULL)
	{
		exitcount++;
	}
	if (exitcount == 0)
	{
		return("");
	}
	exitchoice = lrand48() % exitcount;
	start_traverse_hash(&t,exits);
	tempbucket = next_traverse_hash(&t);
	while (exitchoice > 0)
	{
		exitchoice--;
		tempbucket = next_traverse_hash(&t);
	}
	return(tempbucket->key.text);
}

static void mob_go(struct node *mob, char *dir)
{
	char output_buffer[OUT_BUFFER_SIZE];
	struct node *room, *dest, *exits, *ex;

	room = find_where(mob);
	exits = find_string_node(room,"exits");
	if (exits != (struct node *)NULL)
	{
		ex = find_string_node(exits,dir);
	} else {
		ex = (struct node *)NULL;
	}
	if (ex != (struct node *)NULL)
	{
		if (!(get_named_number(mob,"flags") & SNEAKY_FLAG))
		{
			sprintf(output_buffer,
				"%s leaves %s.\r\n",
				get_named_string(mob,"name"), dir);
			tell_room(find_where(mob), output_buffer,
				mob, (struct node *)NULL, CLASS_MOB);
		}
		dest = find_keyed_node(objects,ex);
		if (dest == (struct node *)NULL)
		{
			sprintf(output_buffer,"Can't go %s -- dead link.\r\n",
					      dir);
			tell_object(mob, output_buffer, CLASS_SELF);
			return;
		}
		insert_object(mob, ex->u.text);
		if (!(get_named_number(mob,"flags") & SNEAKY_FLAG))
		{
			sprintf(output_buffer, "%s arrives.\r\n",
				get_named_string(mob,"name"));
			tell_room(dest, output_buffer,
				mob, (struct node *)NULL, CLASS_MOB);
		}
		sprintf(output_buffer, "You go %s.\r\n", dir);
		tell_object(mob, output_buffer, CLASS_SELF);
		mob_look(mob);
		return;
	}
	sprintf(output_buffer,"Can't go %s.\r\n", dir);
	tell_object(mob, output_buffer, CLASS_SELF);
}

static void do_go(int who, char *command, char *arg)
{
	mob_go(p[who].object, arg);
}

static void do_namedir(int who, char *command, char *arg)
{
	do_go(who,"",command);
}

static int matchname(char *arg, char *name)
{
	return(strstr(name,arg)!=NULL);
}

static void do_kill(int who, char *command, char *arg)
{
	/* TODO */
	char output_buffer[OUT_BUFFER_SIZE];
	struct node *ob, *room, *others;

	ob = p[who].object;
	room = find_where(ob);
	others = find_firstinv(room);
	while (others != (struct node *)NULL)
	{
		if (matchname(arg,get_named_string(others,"name")))
		{
			if (ob == others)
			{
				sprintf(output_buffer,"Killing yourself is sufficiently dumb that it isn't coded yet.\r\n");
				send_string(who, output_buffer, 0, CLASS_SELF);
				return;
			}
			if (matchname("Slimy the snail",get_named_string(others,"name")) ||
			    matchname("Slowly the slime mold",get_named_string(others,"name")))
			{
				sprintf(output_buffer,"You get covered in slime for your efforts.\r\n");
				send_string(who, output_buffer, 0, CLASS_SELF);
				sprintf(output_buffer,
	"%s gets covered in slime attacking %s.\r\n",
					get_named_string(ob,"name"), get_named_string(others,"name"));
				tell_room(room, output_buffer, ob, others, CLASS_MOB);
				sprintf(output_buffer,"%s tries to attack you, and gets covered in slime for their efforts.\r\n",get_named_string(ob,"name"));
				tell_object(others, output_buffer, CLASS_MOB);
				return;
			}
			if (matchname( "Sneaky the ninja",get_named_string(others,"name")))
			{
				sprintf(output_buffer,"%s easily counters your attempt.\r\n",get_named_string(others,"name"));
				send_string(who, output_buffer, 0, CLASS_SELF);
				sprintf(output_buffer,
	"%s futilely tries to attack %s.\r\n",
					get_named_string(ob,"name"), get_named_string(others,"name"));
				tell_room(room, output_buffer, ob, others, CLASS_MOB);
				sprintf(output_buffer,"%s tries to attack you, but you easily counter.\r\n",get_named_string(ob,"name"));
				tell_object(others, output_buffer, CLASS_MOB);
				return;
			}
			sprintf(output_buffer,"You and %s are evenly matched, so your attempt fails.\r\n", get_named_string(others,"name"));
			send_string(who, output_buffer, 0, CLASS_SELF);
			sprintf(output_buffer,
	"%s attacks %s, but they are evenly matched.\r\n",
				get_named_string(ob,"name"), get_named_string(others,"name"));
			tell_room(room, output_buffer, ob, others, CLASS_MOB);
			sprintf(output_buffer,"%s tries to attack you, but you are evenly matched, so they fail.\r\n",get_named_string(ob,"name"));
			tell_object(others, output_buffer, CLASS_MOB);
			return;
		}
		others = find_nextinv(others);
	}
	sprintf(output_buffer,"There is no %s here to kill.\r\n", arg);
	send_string(who, output_buffer, 0, CLASS_SELF);
}

static void do_tell(int who, char *command, char *arg)
{
	/* TODO */
	char string_buffer[OUT_BUFFER_SIZE], output_buffer[OUT_BUFFER_SIZE];
	char tbuffer[80];
	struct tm timestuff;
	int playerdest;
	time_t now;

	now = current_time.tv_sec;
	timestuff = *gmtime(&now);

	if (sscanf(arg,"%d %[^\r\n]",&playerdest,string_buffer) != 2)
	{
		send_string(who,
"That didn't parse as a number followed by a string.\r\n",
			0, CLASS_SELF);
		return;
	}
	if ((playerdest < 0) || (playerdest >= max_players))
	{
		send_string(who, "That player number is out of range.\r\n",
			0, CLASS_SELF);
		return;
	}
	if (p[playerdest].state == -1)
	{
		send_string(who, "No player with that number.\r\n",
			0, CLASS_SELF);
		return;
	}
	if (who == playerdest)
	{
		sprintf(output_buffer,"You tell yourself %s\r\n", string_buffer);
		send_string(who, output_buffer, 0, CLASS_SELF);
		strftime(tbuffer,80,"(at %H:%M:%S) ",&timestuff);
		fprintf(stderr,"%s%s tells %s %s\r\n",tbuffer,get_named_string(p[who].object,"name"),get_named_string(p[playerdest].object,"name"),string_buffer);
		return;
	}
	sprintf(output_buffer,"You tell %s %s\r\n", get_named_string(p[playerdest].object,"name"),string_buffer);
	send_string(who, output_buffer, 0, CLASS_SELF);
	sprintf(output_buffer,"%s tells you %s\r\n", get_named_string(p[who].object,"name"),string_buffer);
	send_string(playerdest, output_buffer, 1, CLASS_OTHER);
	sprintf(output_buffer,"%s tells you %s\r\n", get_named_string(p[who].object,"name"),string_buffer);
	strftime(tbuffer,80,"(at %H:%M:%S) ",&timestuff);
	/* TODO: turn into log call */
	fprintf(stderr,"%s%s tells %s %s\r\n",tbuffer,get_named_string(p[who].object,"name"),get_named_string(p[playerdest].object,"name"),string_buffer);
}

static void do_broadcast(int who, char *command, char *arg)
{
	char output_buffer[OUT_BUFFER_SIZE];

	sprintf(output_buffer,"%s\r\n", arg);
	player_broadcast(output_buffer,1,CLASS_OTHER);
}

static void do_user(int who, char *command, char *arg)
{
	set_named_string(p[who].object,"user",arg);
	send_string(who, "Username set.\r\n", 0, CLASS_SELF);
}

static void do_pass(int who, char *command, char *arg)
{
	char *curuser, *tstring;
	struct node *account;

	curuser = get_named_string(p[who].object,"user");
	account = find_string_node(auth,curuser);
	if ((account != (struct node *)NULL) &&
	    !strcmp(arg,get_named_string(account,"password")))
	{
		tstring = get_named_string(account,"name");
		if (*tstring)
			set_named_string(p[who].object,"name",tstring);
		set_named_number(p[who].object,"priv",
			get_named_number(account,"priv"));
		send_string(who, "Success\r\n", 0, CLASS_SELF);
	} else {
		send_string(who, "Failure\r\n", 0, CLASS_SELF);
	}
}

static void do_dump(int who, char *command, char *arg)
{
	print_hash(&db,stderr,0);
	send_string(who, "DB dumped.\r\n", 0, CLASS_SELF);
}

static void do_loadauth(int who, char *command, char *arg)
{
	FILE *fp;

	fp = fopen("authdb","rb");
	if (fp != (FILE *)NULL)
	{
		read_hash(auth, fp);
		fclose(fp);
	}
	if (who != -1)
		send_string(who, "Auth db reloaded.\r\n", 0, CLASS_SELF);
}

/* flags:
 * 1 = takes 1 arg...
 * -1 = end of list */

/* TODO: move commands into a table (or some of it) */
static int check_commandlist(int who, char *command)
{
	int index, priv;

	index = 0;
	priv = get_named_number(p[who].object,"priv");
	while (commands[index].flags != O_INVALID)
	{
		if (!(commands[index].flags & O_PRIV) || priv)
		{
			if (commands[index].flags & O_ARG)	/* takes arg */
			{
				if (strncmp(command,
					    commands[index].name,
					    commands[index].length) == 0)
				{
					commands[index].func(who, command,
						command+commands[index].length);
					return(1);
				}
			} else {
				if (strcmp(command,commands[index].name) == 0)
				{
					commands[index].func(who,command,"");
					return(1);
				}
			}
		} else {
			if (commands[index].flags & O_ARG)	/* takes arg */
			{
				if (strncmp(command,
					    commands[index].name,
					    commands[index].length) == 0)
				{
					send_string(who, commands[index].name,
						0, CLASS_SELF);
					send_string(who,
						"requires privilige.\r\n",
							0, CLASS_SELF);
					return(1);
				}
			} else {
				if (strcmp(command,commands[index].name) == 0)
				{
					send_string(who, commands[index].name,
						0, CLASS_SELF);
					send_string(who,
						" requires privilige.\r\n",
							0, CLASS_SELF);
					return(1);
				}
			}
		}
		index++;
	}
	return(0);
}

/* This assumes iso 8859-1, but that's pretty safe -- replace with some
 * similar code if you're on an different system */
/* control characters supported:
 * ^H and ^? -- delete prev char.
 * ^U -- back to start of string
 * ^W -- back to last whitespace before last non-whitespace
 * ^I -- tab is legal
 * \r & \n won't show up.
 */
static void sanefilter_string(char *string)
{
	char *in, *out;
	int testchar;

	in = out = string;
	while (*in)
	{
		testchar = (unsigned char)*in;
		/* from space to ~ legal, also NBSP to 255 */
		if (((testchar >= 32) && (testchar < 127)) ||
		     ((testchar >= 160) && (testchar < 256)))
		{
			*out++ = *in++;
		}
		else switch (*in)
		{
			case '\b':	/* backspace */
			case 127:	/* delete */
				if (out > string)
				{
					out--;	/* eat last char */
					in++;
				} else {
					/* otherwise ignore */
					in++;
				}
				break;
			case 21:	/* control-U */
				out = string;
				in++;
				break;
			case 23:	/* control-W */
				while ((out > string) && (isspace(*(out-1))))
				{
					out--;
				}
				while ((out > string) && (!isspace(*(out-1))))
				{
					out--;
				}
				in++;
				break;
			case '\t':	/* tab */
				*out++ = *in++;
				break;
			case '\a':	/* allert = bell = ^G */
				*out++ = *in++;
				break;
			default:
				in++;	/* skip */
				break;
		}
	}
	*out = 0;

}

static void parse_one_command(int who, char *command)
{
	char output_buffer[OUT_BUFFER_SIZE];

	if (p[who].state == 1)	/* chat mode */
	{
		if (strcmp(command,".") == 0)
		{
			p[who].state = 0;
			send_string(who, "Ending chat mode.\r\n",
				0, CLASS_SELF);
		} else {
			do_speech(who,"",command);
		}
	}
	else if (p[who].state > 2)	/* paste mode */
	{
		if (strcmp(command,".") == 0)
		{
			sprintf(output_buffer,"%s says: %s\r\n",
				get_named_string(p[who].object,"name"), p[who].paste_buffer);
			broadcast(output_buffer,1,CLASS_OTHER);
			p[who].state -= 3;
			p[who].paste_buffer[0] = 0;
		}
		else if ((strcmp(command,"@") == 0) ||
			 (strcmp(command,"@abort") == 0))
		{
			send_string(who, "Paste abort.\r\n",
				1,CLASS_SELF);
			p[who].state -= 3;
			p[who].paste_buffer[0] = 0;
		}
		else if (strlen(command) + strlen(p[who].paste_buffer) <
			 PASTE_BUFFER_SIZE-3)
		{
			strcat(p[who].paste_buffer,"\n");
			strcat(p[who].paste_buffer,command);
		} else {
			send_string(who, "Paste buffer overflow.\r\n",
				1,CLASS_SELF);
			p[who].paste_buffer[0] = 0;
		}
	}
	else if (!check_commandlist(who,command))
	{
		/* unrecognized command case */
		if (*command)
		{
			if (p[who].state == 0)
			{
				send_string(who, "Command \"",1,CLASS_SELF);
				send_string(who, command,1,CLASS_SELF);
				send_string(who,
					"\" unrecognized: try \"help\".\r\n",
					1,CLASS_SELF);
			} else {
				do_speech(who,"",command);
			}
		}
	}
}

static void parse_input(int who, char *buffer, int len, int oob)
{
	char *command, *startbuffer;
	char output_buffer[OUT_BUFFER_SIZE];
	int nlflag;

	if (oob)
	{
		sprintf(output_buffer,"%s oob condition: %s\r\n",
			get_named_string(p[who].object,"name"), buffer);
		broadcast(output_buffer,1,CLASS_SYSTEM);
		return;
	}
	startbuffer = buffer;
	while (len > 0)
	{
		if ((*buffer != '\n') && (*buffer != '\r'))
		{
			if (p[who].command_tail < READ_BUFFER_SIZE)
			{
				p[who].command_buffer[p[who].command_tail] =
					*buffer;
				p[who].command_tail++;
			} else {
				/* command too long -- nuke */
				p[who].command_buffer[p[who].command_tail]=0;
				p[who].command_tail = 0;
				send_string(who, "Overlong command aborted.\r\n",1,CLASS_SELF);
				return;
			}
		} else {
			/* terminate the string, if needed */
			p[who].command_buffer[p[who].command_tail]=0;
			/* and remove partial buffer */
			p[who].command_tail = 0;
			command = p[who].command_buffer;
			sanefilter_string(command);
			parse_one_command(who,command);
			startbuffer = buffer;
			nlflag = 1;
			while(((nlflag && (buffer[1] == '\n')) ||
			       (buffer[1] == '\r')) &&
			      (len > 0))
			{
				if (buffer[1] == '\n') nlflag = 0;
				buffer++;
				len--;
			}
		}
		buffer++;
		len--;
	}
	p[who].command_buffer[p[who].command_tail]=0;
}

static void end_player(int who)
{
	char output_buffer[OUT_BUFFER_SIZE];

	/* reinit player slot */
	p[who].command_buffer[0] = 0;
	p[who].paste_buffer[0] = 0;
	p[who].command_tail = 0;
	p[who].state = -1;	/* open slot */
	p[who].wrap = 0;
	sprintf(output_buffer,"Bye, %d!\r\n",who);
	if (p[who].object != (struct node *)NULL)
		free_object(p[who].object);
	p[who].object = (struct node *)NULL;
	broadcast(output_buffer,0,CLASS_SYSTEM);
}

static void mob_ai(struct timeval when, int data)
{
	struct timeval temp;

	mob_go(m[data].object, random_exit(find_where(m[data].object)));
	temp.tv_sec = when.tv_sec + m[data].delay;
	temp.tv_usec = when.tv_usec;
	queue_event(temp,mob_ai,data);
}

static void config_port(int port)
{
	port_number = port;
}

static void config_opt(char *opt, char *val)
{
	long int port;
	char *endptr;

	if (!strcmp(opt,"port"))
	{
		port = strtol(val,&endptr,0);
		/* converted all of val? */
		if ((*endptr == 0) && (endptr != val))
		{
			port_number = port;
		}
	}
}

struct interface bigworld_supplies[] = {
	{ "config_port", config_port, "void(int)", NULL },
	{ "config_opt", config_opt, "void(char*,char*)", NULL },
	{ "new_room", new_room, "void(char*,char*)", NULL },
	{ "link_exit", link_exit, "void(char*,char*,char*)", NULL },
	{NULL,NULL,NULL,NULL}
};

struct interface bigworld_uses[] = {
	{ "queue_event", NULL, "void(timeval,void(*)(timeval,int),int)", NULL},
	{ "close_connection", NULL, "void(int)", NULL },
	{ "real_connection", NULL, "int(int)", NULL },
	{ "connection_name", NULL, "char*(int)", NULL },
	{ "connection_time", NULL, "timeval(int)", NULL },
	{ "open_port", NULL, "int(int,unsigned int,int(*)(int,void(*)(int,char*,int,int)),void(*)(int),void(*)(int,char*,int,int))", NULL },
	{ "close_port", NULL, "void(int)", NULL },
	{ NULL, NULL, NULL, NULL }
};

void bigworld_use_functions(struct interface *it)
{
	struct interface *temp;

	temp = it;
	while (temp->name != NULL)
	{
		if ((!strcmp(temp->name,"queue_event")) &&
		    (!strcmp(temp->prototype,"void(timeval,void(timeval,int),int)")))
		{
			queue_event = temp->function;
		}
		else if ((!strcmp(temp->name,"close_connection")) &&
			 (!strcmp(temp->prototype,"void(int)")))
		{
			close_connection = temp->function;
		}
		else if ((!strcmp(temp->name,"real_connection")) &&
			 (!strcmp(temp->prototype,"int(int)")))
		{
			real_connection = temp->function;
		}
		else if ((!strcmp(temp->name,"connection_name")) &&
			 (!strcmp(temp->prototype,"char*(int)")))
		{
			connection_name = temp->function;
		}
		else if ((!strcmp(temp->name,"connection_time")) &&
			 (!strcmp(temp->prototype,"timeval(int)")))
		{
			connection_time = temp->function;
		}
		else if ((!strcmp(temp->name,"open_port")) &&
			 (!strcmp(temp->prototype,"int(int,unsigned int,int(*)(int,void(*)(int,char*,int,int)),void(*)(int),void(*)(int,char*,int,int))")))
		{
			open_port = temp->function;
		}
		else if ((!strcmp(temp->name,"close_port")) &&
			 (!strcmp(temp->prototype,"void(int)")))
		{
			close_port = temp->function;
		}
		temp++;
	}
	return;
}

int bigworld_initialize(void)
{
	init_assoc(&db,4);
	objects = new_string_node(&db,"objects");
	auth = new_string_node(&db,"auth");
	/* read auth info */
	do_loadauth(-1,NULL,NULL);	/* dummy args */
	obcount = new_string_node(&db,"obcount");
	obcount->type = NUMBER;
	obcount->u.num = 1;
	p = (struct playerinfo *)
		s_malloc(max_players*sizeof(struct playerinfo));
	m = (struct mobinfo *)s_malloc(max_mobs*sizeof(struct mobinfo));
	return(0);
}

/* TODO: move some of this to initialize... */
/* TODO: add support for a go function */
int bigworld_start(void)
{
	int i;
	struct node *slimy, *slowly, *sneaky, *harvey;
	struct timeval temp;

	if ((queue_event == NULL) ||
	    (close_connection == NULL) ||
	    (real_connection == NULL) ||
	    (connection_name == NULL) ||
	    (connection_time == NULL) ||
	    (open_port == NULL) ||
	    (close_port == NULL))
	{
		fprintf(stderr,"Needed functions still uninitialized in bigworld_start!\n");
		return(-1);
	}

	open_port(port_number,INADDR_ANY,new_player,end_player,parse_input);

	root_object = new_object("root object");
	set_named_string(root_object,"name","root_object");

	/* TODO: use a room loader of some sort */
	new_room("start room", "In the start room");
	new_room("other room", "In the other (east) room");
	new_room("south room", "In the south room");
	link_exit("west", "start room", "other room");
	link_exit("east", "other room", "start room");
	link_exit("south", "start room", "south room");
	link_exit("north", "south room", "start room");

	start_room = named_object("start room");
	if (start_room == (struct node *)NULL)
	{
		fprintf(stderr,"Invalid start room!\n");
		return(-1);
	}

	slimy = new_object("slimy");
	set_named_string(slimy,"name","Slimy the snail");
	insert_object(slimy, "other room");
	m[0].object = slimy;
	m[0].delay = 1800;

	slowly = new_object("slowly");
	set_named_string(slowly,"name","Slowly the slime mold");
	insert_object(slowly, "other room");
	m[1].object = slowly;
	m[1].delay = 3600;

	sneaky = new_object("sneaky");
	set_named_string(sneaky,"name","Sneaky the ninja");
	set_named_number(sneaky,"flags",SNEAKY_FLAG);
	insert_object(sneaky, "other room");
	m[2].object = sneaky;
	m[2].delay = 14;

	harvey = new_object("harvey");
	set_named_string(harvey,"name","Harvey the rabbit");
	set_named_number(harvey,"flags",SNEAKY_FLAG | INVIS_FLAG);
	insert_object(harvey, "other room");
	m[3].object = harvey;
	m[3].delay = 120;

	gettimeofday(&current_time,NULL);
	temp.tv_sec = current_time.tv_sec + 900;
	temp.tv_usec = current_time.tv_usec;
	queue_event(temp, mob_ai, 0);
	temp.tv_sec = current_time.tv_sec + 1800;
	temp.tv_usec = current_time.tv_usec;
	queue_event(temp, mob_ai, 1);
	temp.tv_sec = current_time.tv_sec + 7;
	temp.tv_usec = current_time.tv_usec;
	queue_event(temp, mob_ai, 2);
	temp.tv_sec = current_time.tv_sec + 60;
	temp.tv_usec = current_time.tv_usec;
	queue_event(temp, mob_ai, 3);

	/* initialize player data to known values */
	for(i=0;i<max_players;i++)
	{
		p[i].command_buffer[0] = 0;
		p[i].command_tail = 0;
		p[i].paste_buffer[0] = 0;
		p[i].state = -1;	/* open slot */
		p[i].wrap = 0;
	}
	return(0);
}

/* TODO: more carefully thought out shutdown options */
int bigworld_stop(int severity)
{
	/* TODO: severity 0: ask.  1: Shutdown!  2: 'recovering' from fatal */
	/* TODO: mud closure stuff (close socket, wait for no players..) */
	/* TODO: dump the db */
	if (severity > 0)
	{
		cleanup_assoc(&db);
		return(0);
	} else {
		return(-1);
	}
}

/* TODO: line-by-line check & cleanup */
/* TODO: fix naming convention */
/* TODO: split into modules */
/* TODO: something more extensible than simple object flags */
/* TODO: fail checks -- bail in assoc functions */
/* TODO: remove unnecessary fail checks */
/* TODO: central panic function in loader, etc */
/* TODO: linked-list mode for small hashes? */
/* TODO: fd counter in select module */
/* TODO: fileconfig module */
/* TODO: remove 'magic' constants */
/* TODO: eline preference */
/* TODO: sayto command */
