/* Example DevMUD module to exercise interface standard and socket module,
 * using a tiled graphical client */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include "../include/devmud.h"

struct event {
	struct timeval when;
	struct object *ob;
	int data;
};

struct heap {
	struct event *events;
	int num_events, count;
} h;

struct object {
	int x, y, dir;
};

static void delete_event(void);

/* TODO intern strings */

/* TODO static number of players is very bad to assume */
#define MAX_PLAYERS 200 /* only 256 descriptors... */
#define READ_BUFFER_SIZE 16384
#define OUT_BUFFER_SIZE (READ_BUFFER_SIZE+256)
#define LASTLOG_SIZE 10

static void (*close_connection)(int playerid) = NULL;
static void (*parsed_put_string)(int to, char *message) = NULL;
static int (*get_message)(char *buffer, int size, struct timeval *delay) = NULL;
static int (*real_connection)(int playerid) = NULL;
static char *(*connection_name)(int playerid) = NULL;
static int (*connection_time)(int playerid) = NULL;

struct playerinfo {
	char command_buffer[READ_BUFFER_SIZE];
	int command_tail;
	int x, y;
} p[MAX_PLAYERS];

#define MAPX 10
#define MAPY 10

static int map[MAPX][MAPY] = {
	{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
	{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
	{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
	{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
	{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
	{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
	{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
	{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
	{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
};

static void queue_event(struct timeval when, struct object *ob, int data)
{
	unsigned int loc, newloc;
	struct event tempevent, *new_events;

	if (h.count == h.num_events)
	{
		h.num_events *= 2;

		/* get more memory */
		new_events = (struct event *)
			realloc(h.events,(h.num_events)*sizeof(struct event));

		/* and check it */
		if (new_events == (struct event *)NULL)
		{
			fprintf(stderr,"Couldn't realloc event events.\n");
			/* put it back... */
			h.num_events /= 2;
			exit(1);
		}
		h.events = new_events;
	}
	loc = h.count++;
	h.events[loc].when = when;
	h.events[loc].ob = ob;
	h.events[loc].data = data;
	/* event at loc may now violate the heap property... */

	newloc = (loc-1)/2;
	while ((loc > 0) &&
	       ((h.events[loc].when.tv_sec < h.events[newloc].when.tv_sec) ||
		((h.events[loc].when.tv_sec == h.events[newloc].when.tv_sec) &&
		 (h.events[loc].when.tv_sec < h.events[newloc].when.tv_sec))))
	{
		tempevent = h.events[loc];
		h.events[loc] = h.events[newloc];
		h.events[newloc] = tempevent;
		loc = newloc;
		newloc = (loc-1)/2;
	}
}

static void send_string(int playerid, char *str)
{
	parsed_put_string(playerid,str);
}

static void tile_announce(int x, int y)
{
	int loop;
	char buffer[80];

	sprintf(buffer,"Tile %d,%d is %d\n",x,y,map[x][y]);

	for (loop=0;loop<MAX_PLAYERS;loop++)
	{
		if (real_connection(loop))
		{
			send_string(loop,buffer);
		}
	}
}

static void new_player(int playerid)
{
	int x, y;
	char buffer[80];

	p[playerid].command_tail = 0;
	/* find an empty space to put them */
	p[playerid].x = 8;
	p[playerid].y = 8;
	for (x = 1; (x<MAPX) && (p[playerid].x == 8); x++)
	{
		for (y = 1; (y<MAPX) && (p[playerid].x == 8); y++)
		{
			if (map[x][y] == 0)
			{
				p[playerid].x = x;
				p[playerid].y = y;
				map[x][y] = 2;
				tile_announce(x,y);
			}
		}
	}
	/* and report the whole map */
	for (x = 0; x<MAPX; x++)
	{
		for (y = 0; y<MAPX; y++)
		{
			sprintf(buffer,"Tile %d,%d is %d\n",x,y,map[x][y]);
			send_string(playerid,buffer);
		}
	}
}

static int move_tile(int x, int y, int dx, int dy)
{
	/* legal move? */
	if (map[x+dx][y+dy] == 0)
	{
		map[x+dx][y+dy] = map[x][y];
		map[x][y] = 0;
		tile_announce(x,y);
		tile_announce(x+dx,y+dy);
		return(1);	/* happened */
	}
	return(0);	/* did not happen */
}

static int check_commandlist(int playerid, char *command)
{
	int index;

	index = 0;
	if (strncmp(command,"north",5) == 0)
	{
		if (move_tile(p[playerid].x,p[playerid].y,0,-1))
		{
			p[playerid].y -= 1;
		}
		return(1);
	}
	if (strncmp(command,"south",5) == 0)
	{
		if (move_tile(p[playerid].x,p[playerid].y,0,1))
		{
			p[playerid].y += 1;
		}
		return(1);
	}
	if (strncmp(command,"east",4) == 0)
	{
		if (move_tile(p[playerid].x,p[playerid].y,1,0))
		{
			p[playerid].x += 1;
		}
		return(1);
	}
	if (strncmp(command,"west",4) == 0)
	{
		if (move_tile(p[playerid].x,p[playerid].y,-1,0))
		{
			p[playerid].x -= 1;
		}
		return(1);
	}
	if (strncmp(command,"quit",4) == 0)
	{
		close_connection(p[playerid].id);
		return(1);
	}
	return(0);
}

static void parse_input(int playerid, char *buffer)
{
	char *command;
	char output_buffer[OUT_BUFFER_SIZE];

	/* read until newline, or... */
	while ((*buffer != '\n') && (*buffer != '\r') &&
	/* we really filled up the buffer. */
	       (p[playerid].command_tail < READ_BUFFER_SIZE))
	{
		p[playerid].command_buffer[p[playerid].command_tail] = *buffer;
		if (*buffer == 0)
		{
			return;
		}
		p[playerid].command_tail++;
		buffer++;
	}
	/* terminate the string, if needed */
	p[playerid].command_buffer[p[playerid].command_tail] = 0;
	/* and inform the next time that there isn't a partial buffer */
	p[playerid].command_tail = 0;
	command = p[playerid].command_buffer;

	if (!check_commandlist(playerid,command))
	{
		/* unrecognized command case */
		if (*command)
		{
			sprintf(output_buffer,"Command \"%s\" unrecognized: try \"help\".\r\n",command);
			send_string(playerid, output_buffer);
		}
	}
}

static void end_player(int playerid)
{
	map[p[playerid].x][p[playerid].y] = 0;
	tile_announce(p[playerid].x,p[playerid].y);
}

static void mob_ai(struct timeval when, struct object *mob, int data)
{
	int action;
	struct timeval temp;

	action = (lrand48()%8);
	switch(action)
	{
		case 0:
			mob->dir = ((mob->dir+1) % 4);
			break;
		case 1:
			mob->dir = ((mob->dir+2) % 4);
			break;
		case 2:
			mob->dir = ((mob->dir+3) % 4);
			break;
		default:
			switch(mob->dir)
			{
				case 0:
					if (move_tile(mob->x,mob->y,0,-1))
					{
						mob->y--;
					}
					break;
				case 1:
					if (move_tile(mob->x,mob->y,1,0))
					{
						mob->x++;
					}
					break;
				case 2:
					if (move_tile(mob->x,mob->y,0,1))
					{
						mob->y++;
					}
					break;
				case 3:
					if (move_tile(mob->x,mob->y,-1,0))
					{
						mob->x--;
					}
					break;
			}
			break;
	}
	temp.tv_sec = when.tv_sec + data;
	temp.tv_usec = when.tv_usec;
	queue_event(temp,mob,data);
}

/* remove the top heap event */
static void delete_event(void)
{
	struct event tempevent;
	int loc, newloc, end;

	if (h.count == 0)
	{
		/* heap is empty -- can't get an event */
		return;
	}

	h.count--;
	end = h.count;		/* last valid spot */
	/* stuff the last event at the top -- we'll fix it shortly */
	h.events[0] = h.events[end];
	end--;			/* that spot is no longer valid */
	loc = 0;

	while (1)	/* return from the loop when it's ok */
	{
		newloc = (loc * 2) + 1;
		if (newloc >= end)		/* < 2 children */
		{
			if ((newloc == end) &&
			    ((h.events[newloc].when.tv_sec <
					h.events[loc].when.tv_sec) ||
			     ((h.events[newloc].when.tv_sec ==
					h.events[loc].when.tv_sec) &&
			      (h.events[newloc].when.tv_usec <
					h.events[loc].when.tv_usec))))

			{
				tempevent = h.events[loc];
				h.events[loc] = h.events[newloc];
				h.events[newloc] = tempevent;
			}
			return;
		}
		/* now known to have 2 children */
		if ((h.events[newloc].when.tv_sec > h.events[newloc+1].when.tv_sec) ||
		    ((h.events[newloc].when.tv_sec == h.events[newloc+1].when.tv_sec) &&
		     (h.events[newloc].when.tv_usec > h.events[newloc+1].when.tv_usec)))

		{
			newloc++;
		}
		if ((h.events[newloc].when.tv_sec <
				h.events[loc].when.tv_sec) ||
		    ((h.events[newloc].when.tv_sec ==
				h.events[loc].when.tv_sec) &&
		     (h.events[newloc].when.tv_usec <
				h.events[loc].when.tv_usec)))
		{
				tempevent = h.events[loc];
				h.events[loc] = h.events[newloc];
				h.events[newloc] = tempevent;
				loc = newloc;
		} else {
			return;
		}
	}
}

struct interface tileworld_supplies[] = {
	{NULL,NULL,NULL,NULL}
};

struct interface tileworld_uses[] = {
	{ "close_connection", NULL, "void(int)", NULL },
	{ "parsed_put_string", NULL, "void(int,char*)", NULL },
	{ "get_message", NULL, "int(char*,int,timeval*)", NULL },
	{ "real_connection", NULL, "int(int)", NULL },
	{ "connection_name", NULL, "char*(int)", NULL },
	{ "connection_time", NULL, "int(int)", NULL },
	{ NULL, NULL, NULL, NULL }
};

void tileworld_use_functions(struct interface *it)
{
	struct interface *temp;

	temp = it;
	while (temp->name != NULL)
	{
		if ((!strcmp(temp->name,"close_connection")) &&
		    (!strcmp(temp->prototype,"void(int)")))
		{
			close_connection = temp->function;
		}
		else if ((!strcmp(temp->name,"parsed_put_string")) &&
			 (!strcmp(temp->prototype,"void(int,char*)")))
		{
			parsed_put_string = temp->function;
		}
		else if ((!strcmp(temp->name,"get_message")) &&
			 (!strcmp(temp->prototype,"int(char*,int,timeval*)")))
		{
			get_message = temp->function;
		}
		else if ((!strcmp(temp->name,"real_connection")) &&
			 (!strcmp(temp->prototype,"int(int)")))
		{
			real_connection = temp->function;
		}
		else if ((!strcmp(temp->name,"connection_name")) &&
			 (!strcmp(temp->prototype,"char*(int)")))
		{
			connection_name = temp->function;
		}
		else if ((!strcmp(temp->name,"connection_time")) &&
			 (!strcmp(temp->prototype,"int(int)")))
		{
			connection_time = temp->function;
		}
		temp++;
	}
	return;
}

int tileworld_initialize(void)
{
	return(0);
}

/* TODO: move some of this to initialize... */
/* TODO: add support for a go function */
/* TODO: split into modules */
int tileworld_start(void)
{
	char read_buffer[READ_BUFFER_SIZE];
	int i, playerid;
	/* TODO: use a room loader of some sort */
	struct object slimy;
	struct timeval now, temp, delay;
	struct event current;

	if ((close_connection == NULL) ||   
	    (parsed_put_string == NULL) ||
	    (get_message == NULL) ||
	    (real_connection == NULL) ||
	    (connection_name == NULL) ||
	    (connection_time == NULL))
	{
		fprintf(stderr,"Needed functions still uninitialized in tileworld_start!\n");
		return(-1);
	}

	/* set up the heap for events */
	h.count = 0;
	h.num_events = 32;	/* arbitrary number of starting events */
	/* get the memory */
	h.events = (struct event *)
			malloc((h.num_events)*sizeof(struct event));
	/* and check it */
	if (h.events == (struct event *)NULL)
	{
		fprintf(stderr,"Couldn't allocate memory for heap events.\n");
		return(-1);
	}

	slimy.x = 4;
	slimy.y = 4;
	slimy.dir = 0;
	map[slimy.x][slimy.y] = 3;
	gettimeofday(&now,NULL);
	temp.tv_sec = now.tv_sec + 2;
	temp.tv_usec = now.tv_usec;
	queue_event(temp, &slimy, 2);

	/* initialize player data to known values */
	for(i=0;i<MAX_PLAYERS;i++)
	{
		p[i].command_buffer[0] = 0;
		p[i].command_tail = 0;
		p[i].x = 0;
		p[i].y = 0;
	}
	while(1)
	{
		read_buffer[0] = 0;
		gettimeofday(&now,NULL);
		if (h.count == 0)
		{
			playerid = get_message(read_buffer,
				READ_BUFFER_SIZE, NULL);
		} else {
			delay.tv_sec = h.events[0].when.tv_sec - now.tv_sec;
			delay.tv_usec = h.events[0].when.tv_usec - now.tv_usec;
			if (delay.tv_usec < 0)  /* normalize */
			{
				delay.tv_usec += 1000000;
				delay.tv_sec--;
			}
			if ((delay.tv_sec > 0) || (delay.tv_usec > 0))
			{
				playerid = get_message(read_buffer,READ_BUFFER_SIZE, &delay);
			} else {
				playerid = -1;
			}
		}
		if (playerid != -1)
		{
			if (*read_buffer)
			{
				parse_input(playerid, read_buffer);
			} else {
				if (real_connection(playerid))
					new_player(playerid);
				else
					end_player(playerid);
			}
		} else if ((h.count != 0) &&
			  ((now.tv_sec > h.events[0].when.tv_sec) ||
			   ((now.tv_sec == h.events[0].when.tv_sec) &&
			    (now.tv_usec >= h.events[0].when.tv_usec)))) {
			current = h.events[0];
			delete_event();
			mob_ai(current.when,current.ob,current.data);
		}
	}
	/* clean up the heap */
	free(h.events);
	return(0);
}

int tileworld_stop(void)
{
	return(-1);
}
